#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 1.0-alpha Toolbox
Â© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0

Updated 5/11/2022 with PH edits
"""

import os
import scipy
import pickle
import numpy as np
from pathlib import Path
import matplotlib.pyplot as plt
from vame.util.auxiliary import read_config
from vame.analysis.tree_hierarchy import graph_to_tree, draw_tree, traverse_tree_cutline
from vame.util.data_manipulation import consecutive
from typing import List, Tuple
from vame.schemas.states import save_state, CommunityFunctionSchema
from vame.logging.logger import VameLogger
from vame.schemas.project import Parametrizations


logger_config = VameLogger(__name__)
logger = logger_config.logger


def get_adjacency_matrix(labels: np.ndarray, n_cluster: int) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Calculate the adjacency matrix, transition matrix, and temporal matrix.

    Args:
        labels (np.ndarray): Array of cluster labels.
        n_cluster (int): Number of clusters.

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: Tuple containing adjacency matrix, transition matrix, and temporal matrix.
    """
    temp_matrix = np.zeros((n_cluster,n_cluster), dtype=np.float64)
    adjacency_matrix = np.zeros((n_cluster,n_cluster), dtype=np.float64)
    cntMat = np.zeros((n_cluster))
    steps = len(labels)

    for i in range(n_cluster):
        for k in range(steps-1):
            idx = labels[k]
            if idx == i:
                idx2 = labels[k+1]
                if idx == idx2:
                    continue
                else:
                    cntMat[idx2] = cntMat[idx2] +1
        temp_matrix[i] = cntMat
        cntMat = np.zeros((n_cluster))

    for k in range(steps-1):
        idx = labels[k]
        idx2 = labels[k+1]
        if idx == idx2:
            continue
        adjacency_matrix[idx,idx2] = 1
        adjacency_matrix[idx2,idx] = 1

    transition_matrix = get_transition_matrix(temp_matrix)

    return adjacency_matrix, transition_matrix, temp_matrix


def get_transition_matrix(adjacency_matrix: np.ndarray, threshold: float = 0.0) -> np.ndarray:
    """Compute the transition matrix from the adjacency matrix.

    Args:
        adjacency_matrix (np.ndarray): Adjacency matrix.
        threshold (float, optional): Threshold for considering transitions. Defaults to 0.0.

    Returns:
        np.ndarray: Transition matrix.
    """
    row_sum=adjacency_matrix.sum(axis=1)
    transition_matrix = adjacency_matrix/row_sum[:,np.newaxis]
    transition_matrix[transition_matrix <= threshold] = 0
    if np.any(np.isnan(transition_matrix)):
        transition_matrix=np.nan_to_num(transition_matrix)
    return transition_matrix



def find_zero_labels(motif_usage: Tuple[np.ndarray, np.ndarray], n_cluster: int) -> np.ndarray:
    """Find zero labels in motif usage and fill them.

    Args:
        motif_usage (Tuple[np.ndarray, np.ndarray]): 2D list where the first index is a unique list of motif used and the second index is the motif usage in frames.
        n_cluster (int): Number of clusters.

    Returns:
        np.ndarray: List of motif usage frames with 0's where motifs weren't used (array with zero labels filled).
    """
    cons = consecutive(motif_usage[0])
    usage_list = list(motif_usage[1])
    if len(cons) != 1: #if missing motif is in the middle of the list
        logger.info("Go")
        if 0 not in cons[0]:
            first_id = cons[0][0]
            for k in range(first_id):
                usage_list.insert(k,0)

        for i in range(len(cons)-1):
            a = cons[i+1][0]
            b = cons[i][-1]
            d = (a-b)-1
            for j in range(1,d+1):
                index = cons[i][-1]+j
                usage_list.insert(index,0)
        if len(usage_list) < n_cluster:
            usage_list.insert(n_cluster,0)

    elif len(cons[0]) != n_cluster: #if missing motif is at the front or end of list
        # diff = n_cluster - cons[0][-1]
        usage_list = list(motif_usage[1])
        if cons[0][0] != 0: #missing motif at front of list
            usage_list.insert(0,0)
        else: #missing motif at end of list
            usage_list.insert(n_cluster-1,0)

    if len(usage_list) < n_cluster: #if there's more than one motif missing
        for k in range(len(usage_list), n_cluster):
            usage_list.insert(k,0)

    usage = np.array(usage_list)
    return usage

def augment_motif_timeseries(label: np.ndarray, n_cluster: int) -> Tuple[np.ndarray, np.ndarray]:
    """Augment motif time series by filling zero motifs.

    Args:
        label (np.ndarray): Original label array.
        n_cluster (int): Number of clusters.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Augmented label array and indices of zero motifs.
    """
    augmented_label = label.copy()

    motif_usage = np.unique(augmented_label, return_counts=True)
    augmented_usage = find_zero_labels(motif_usage, n_cluster)
    zero_motifs = np.where(augmented_usage == 0)[0]
    logger.info(f"Zero motifs: {zero_motifs}")

    idx = -1
    for i in range(len(zero_motifs)):
        for j in range(20):
            x = np.random.choice(zero_motifs)
            augmented_label[idx] = x
            idx -= 1

    return augmented_label, zero_motifs

def get_labels(cfg: dict, files: List[str], model_name: str, n_cluster: int, parametrization: str) -> List[np.ndarray]:
    """Get cluster labels for given videos files.

    Args:
        cfg (dict): Configuration parameters.
        files (List[str]): List of video files paths.
        model_name (str): Model name.
        n_cluster (int): Number of clusters.
        parametrization (str): parametrization.

    Returns:
        List[np.ndarray]: List of cluster labels for each file.
    """
    shapes = []
    for file in files:
        path_to_file = os.path.join(cfg['project_path'],"results",file,'VAME', parametrization + '-' + str(n_cluster),"")
        label = np.load(os.path.join(path_to_file,str(n_cluster)+'_' + parametrization + '_label_'+file+'.npy'))
        shape = len(label)
        shapes.append(shape)
    shapes = np.array(shapes)
    min_frames = min(shapes)

    labels = []
    for file in files:
        path_to_file = os.path.join(cfg['project_path'],"results",file,model_name, parametrization + '-'+str(n_cluster),"")
        label = np.load(os.path.join(path_to_file,str(n_cluster)+'_' + parametrization + '_label_'+file+'.npy'))
        label = label[:min_frames]
        augmented_label,zero_motifs = augment_motif_timeseries(label, n_cluster)
        labels.append(augmented_label)
    return labels

def get_community_label(cfg: dict, files: List[str], model_name: str, n_cluster: int, parametrization: str) -> np.ndarray:
    """Get community labels for given files.

    Args:
        cfg (dict): Configuration parameters.
        files (List[str]): List of files paths.
        model_name (str): Model name.
        n_cluster (int): Number of clusters.
        parametrization (str): parametrization.

    Returns:
        np.ndarray: Array of community labels.
    """
    shapes = []
    for file in files:
        path_to_file = os.path.join(cfg['project_path'],"results",file,'VAME', parametrization + '-'+str(n_cluster),"")
        label = np.load(os.path.join(path_to_file,str(n_cluster)+'_' + parametrization + '_label_'+file+'.npy'))
        shape = len(label)
        shapes.append(shape)
    shapes = np.array(shapes)
    min_frames = min(shapes)

    community_label = []
    for file in files:
        path_to_file = os.path.join(cfg['project_path'],"results",file,'VAME', parametrization + '-'+str(n_cluster),"")
        label = np.load(os.path.join(path_to_file,str(n_cluster)+'_' + parametrization + '_label_'+file+'.npy'))[:min_frames]
        community_label.extend(label)
    community_label = np.array(community_label)
    return community_label


def compute_transition_matrices(files: List[str], labels: List[np.ndarray], n_cluster: int) -> List[np.ndarray]:
    """Compute transition matrices for given files and labels.

    Args:
        files (List[str]): List of file paths.
        labels (List[np.ndarray]): List of label arrays.
        n_cluster (int): Number of clusters.

    Returns:
        List[np.ndarray]: List of transition matrices.
    """
    transition_matrices = []
    for i, file in enumerate(files):
        adj, trans, mat = get_adjacency_matrix(labels[i], n_cluster)
        transition_matrices.append(trans)
    return transition_matrices

def create_community_bag(
    files: List[str],
    labels: List[np.ndarray],
    transition_matrices: List[np.ndarray],
    cut_tree: int,
    n_cluster: int
) -> Tuple:
    """Create community bag for given files and labels (Markov chain to tree -> community detection).
    Args:
        files (List[str]): List of file paths.
        labels (List[np.ndarray]): List of label arrays.
        transition_matrices (List[np.ndarray]): List of transition matrices.
        cut_tree (int): Cut line for tree.
        n_cluster (int): Number of clusters.

    Returns:
        Tuple: Tuple containing list of community bags and list of trees.
    """
    trees = []
    communities_all = []
    for i, file in enumerate(files):
        _, usage = np.unique(labels[i], return_counts=True)
        T = graph_to_tree(usage, transition_matrices[i], n_cluster, merge_sel=1)
        trees.append(T)

        if cut_tree is not None:
            community_bag =  traverse_tree_cutline(T,cutline=cut_tree)
            communities_all.append(community_bag)
            draw_tree(T)
        else:
            draw_tree(T)
            plt.pause(0.5)
            flag_1 = 'no'
            while flag_1 == 'no':
                cutline = int(input("Where do you want to cut the Tree? 0/1/2/3/..."))
                community_bag =  traverse_tree_cutline(T,cutline=cutline)
                logger.info(community_bag)
                flag_2 = input('\nAre all motifs in the list? (yes/no/restart)')
                if flag_2 == 'no':
                    while flag_2 == 'no':
                        add = input('Extend list or add in the end? (ext/end)')
                        if add == "ext":
                            motif_idx = int(input('Which motif number? '))
                            list_idx = int(input('At which position in the list? (pythonic indexing starts at 0) '))
                            community_bag[list_idx].append(motif_idx)
                        if add == "end":
                            motif_idx = int(input('Which motif number? '))
                            community_bag.append([motif_idx])
                        logger.info(community_bag)
                        flag_2 = input('\nAre all motifs in the list? (yes/no/restart)')
                if flag_2 == "restart":
                    continue
                if flag_2 == 'yes':
                    communities_all.append(community_bag)
                    flag_1 = 'yes'

    return communities_all, trees

def create_cohort_community_bag(
    labels: List[np.ndarray],
    trans_mat_full: np.ndarray,
    cut_tree: int,
    n_cluster: int
) -> Tuple:
    """Create cohort community bag for given labels, transition matrix, cut tree, and number of clusters.
    (markov chain to tree -> community detection)

    Args:
        labels (List[np.ndarray]): List of label arrays.
        trans_mat_full (np.ndarray): Full transition matrix.
        cut_tree (int): Cut line for tree.
        n_cluster (int): Number of clusters.

    Returns:
        Tuple: Tuple containing list of community bags and list of trees.
    """

    trees = []
    communities_all = []
    #for i, file in enumerate(files):
    _, usage_full = np.unique(labels, return_counts=True)
    T = graph_to_tree(usage_full, trans_mat_full, n_cluster, merge_sel=1)
    # nx.write_gpickle(T, 'T.gpickle')
    trees.append(T)

    if cut_tree is not None:
        community_bag =  traverse_tree_cutline(T,cutline=cut_tree)
        communities_all = community_bag
        draw_tree(T)
    else:
        draw_tree(T)
        plt.pause(0.5)
        flag_1 = 'no'
        while flag_1 == 'no':
            cutline = int(input("Where do you want to cut the Tree? 0/1/2/3/..."))
            community_bag =  traverse_tree_cutline(T,cutline=cutline)
            logger.info(community_bag)
            flag_2 = input('\nAre all motifs in the list? (yes/no/restart)')
            if flag_2 == 'no':
                while flag_2 == 'no':
                    add = input('Extend list or add in the end? (ext/end)')
                    if add == "ext":
                        motif_idx = int(input('Which motif number? '))
                        list_idx = int(input('At which position in the list? (pythonic indexing starts at 0) '))
                        community_bag[list_idx].append(motif_idx)
                    if add == "end":
                        motif_idx = int(input('Which motif number? '))
                        community_bag.append([motif_idx])
                        logger.info(community_bag)
                    flag_2 = input('\nAre all motifs in the list? (yes/no/restart)')
            if flag_2 == "restart":
                continue
            if flag_2 == 'yes':
                communities_all=community_bag
                flag_1 = 'yes'
    return communities_all, trees

def get_community_labels(files: List[str], labels: List[np.ndarray], communities_all: List[List[List[int]]]) -> List[np.ndarray]:
    """Transform kmeans parameterized latent vector into communities. Get community labels for given files and community bags.

    Args:
        files (List[str]): List of file paths.
        labels (List[np.ndarray]): List of label arrays.
        communities_all (List[List[List[int]]]): List of community bags.

    Returns:
        List[np.ndarray]: List of community labels for each file.
    """
    community_labels_all = []
    for k, file in enumerate(files):
        num_comm = len(communities_all[k])

        community_labels = np.zeros_like(labels[k])
        for i in range(num_comm):
            clust = np.array(communities_all[k][i])
            for j in range(len(clust)):
                find_clust = np.where(labels[k] == clust[j])[0]
                community_labels[find_clust] = i

        community_labels = np.int64(scipy.signal.medfilt(community_labels, 7))
        community_labels_all.append(community_labels)

    return community_labels_all

def get_cohort_community_labels(
    files: List[str],
    labels: List[np.ndarray],
    communities_all: List[List[List[int]]]
) -> List[np.ndarray]:
    """Transform kmeans parameterized latent vector into communities. Get cohort community labels for given labels, and community bags.

    Args:
        files (List[str], deprecated): List of file paths.
        labels (List[np.ndarray]): List of label arrays.
        communities_all (List[List[List[int]]]): List of community bags.

    Returns:
        List[np.ndarray]: List of cohort community labels for each file.
    """
    community_labels_all = []

    num_comm = len(communities_all)

    community_labels = np.zeros_like(labels)
    for i in range(num_comm):
        clust = np.asarray(communities_all[i])

        for j in range(len(clust)):
            find_clust = np.where(labels == clust[j])[0]
            community_labels[find_clust] = i

    community_labels = np.int64(scipy.signal.medfilt(community_labels, 7))
    community_labels_all.append(community_labels)

    return community_labels_all


@save_state(model=CommunityFunctionSchema)
def community(
    config: str,
    parametrization: Parametrizations,
    cohort: bool = True,
    cut_tree: int | None = None,
    save_logs: bool = False
) -> None:
    """Perform community analysis.

    Args:
        config (str): Path to the configuration file.
        cohort (bool, optional): Flag indicating cohort analysis. Defaults to True.
        cut_tree (int, optional): Cut line for tree. Defaults to None.

    Returns:
        None
    """
    try:
        config_file = Path(config).resolve()
        cfg = read_config(config_file)
        parametrizations = cfg['parametrizations']

        if save_logs:
            log_path = Path(cfg['project_path']) / 'logs' / 'community.log'
            logger_config.add_file_handler(log_path)
        model_name = cfg['model_name']
        n_cluster = cfg['n_cluster']

        files = []
        if cfg['all_data'] == 'No':
            all_flag = input("Do you want to write motif videos for your entire dataset? \n"
                        "If you only want to use a specific dataset type filename: \n"
                        "yes/no/filename ")
        else:
            all_flag = 'yes'

        if all_flag == 'yes' or all_flag == 'Yes':
            for file in cfg['video_sets']:
                files.append(file)

        elif all_flag == 'no' or all_flag == 'No':
            for file in cfg['video_sets']:
                use_file = input("Do you want to quantify " + file + "? yes/no: ")
                if use_file == 'yes':
                    files.append(file)
                if use_file == 'no':
                    continue
        else:
            files.append(all_flag)

        if cohort:
            path_to_file = Path(os.path.join(cfg['project_path'], "results", 'community_cohort', parametrization + '-'+str(n_cluster)))

            if not path_to_file.exists():
                path_to_file.mkdir(parents=True, exist_ok=True)

            labels = get_community_label(cfg, files, model_name, n_cluster, parametrization)
            augmented_label, zero_motifs = augment_motif_timeseries(labels, n_cluster)
            _, trans_mat_full,_ = get_adjacency_matrix(augmented_label, n_cluster=n_cluster)
            _, usage_full = np.unique(augmented_label, return_counts=True)
            communities_all, trees = create_cohort_community_bag(labels, trans_mat_full, cut_tree, n_cluster)

            community_labels_all = get_cohort_community_labels(files, labels, communities_all)
            # community_bag = traverse_tree_cutline(trees, cutline=cut_tree)

            # convert communities_all to dtype object numpy array because communities_all is an inhomogeneous list
            communities_all = np.array(communities_all, dtype=object)

            np.save(os.path.join(path_to_file,"cohort_transition_matrix"+'.npy'),trans_mat_full)
            np.save(os.path.join(path_to_file,"cohort_community_label"+'.npy'), community_labels_all)
            np.save(os.path.join(path_to_file,"cohort_" + parametrization + "_label"+'.npy'), labels)
            np.save(os.path.join(path_to_file,"cohort_community_bag"+'.npy'), communities_all)

            with open(os.path.join(path_to_file, "hierarchy"+".pkl"), "wb") as fp:   #Pickling
                pickle.dump(communities_all, fp)

        # Work in Progress
        elif not cohort:
            labels = get_labels(cfg, files, model_name, n_cluster, parametrization)
            transition_matrices = compute_transition_matrices(files, labels, n_cluster)
            communities_all, trees = create_community_bag(files, labels, transition_matrices, cut_tree, n_cluster)
            community_labels_all = get_community_labels(files, labels, communities_all)

            for idx, file in enumerate(files):
                path_to_file=os.path.join(cfg['project_path'],"results",file,model_name, parametrization + '-'+str(n_cluster),"")
                if not os.path.exists(os.path.join(path_to_file,"community")):
                    os.mkdir(os.path.join(path_to_file,"community"))

                np.save(os.path.join(path_to_file,"community","transition_matrix_"+file+'.npy'),transition_matrices[idx])
                np.save(os.path.join(path_to_file,"community","community_label_"+file+'.npy'), community_labels_all[idx])

                with open(os.path.join(path_to_file,"community","hierarchy"+file+".pkl"), "wb") as fp:   #Pickling
                    pickle.dump(communities_all[idx], fp)

    except Exception as e:
        logger.exception(f"Error in community_analysis: {e}")
        raise e
    finally:
        logger_config.remove_file_handler()








#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 1.0-alpha Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""

import os
import h5py
import tqdm
import scipy
import cv2 as cv
import numpy as np
import pandas as pd
from vame.logging.logger import VameLogger
from vame.util.data_manipulation import (
    interpol_first_rows_nans,
    crop_and_flip,
    background
)


logger_config = VameLogger(__name__)
logger = logger_config.logger

def get_animal_frames(
    cfg: dict,
    filename: str,
    pose_ref_index: list,
    start: int,
    length: int,
    subtract_background: bool,
    file_format: str = '.mp4',
    crop_size: tuple = (300, 300)
) -> list:
    """
    Extracts frames of an animal from a video file and returns them as a list.

    Args:
        cfg (dict): Configuration dictionary containing project information.
        filename (str): Name of the video file.
        pose_ref_index (list): List of reference coordinate indices for alignment.
        start (int): Starting frame index.
        length (int): Number of frames to extract.
        subtract_background (bool): Whether to subtract background or not.
        file_format (str, optional): Format of the video file. Defaults to '.mp4'.
        crop_size (tuple, optional): Size of the cropped area. Defaults to (300, 300).

    Returns:
        list: List of extracted frames.
    """
    path_to_file = cfg['project_path']
    time_window = cfg['time_window']
    lag = int(time_window / 2)
    #read out data
    data = pd.read_csv(os.path.join(path_to_file,"videos","pose_estimation",filename+'.csv'), skiprows = 2)
    data_mat = pd.DataFrame.to_numpy(data)
    data_mat = data_mat[:,1:]

    # get the coordinates for alignment from data table
    pose_list = []

    for i in range(int(data_mat.shape[1]/3)):
        pose_list.append(data_mat[:,i*3:(i+1)*3])

    #list of reference coordinate indices for alignment
    #0: snout, 1: forehand_left, 2: forehand_right,
    #3: hindleft, 4: hindright, 5: tail

    pose_ref_index = pose_ref_index

    #list of 2 reference coordinate indices for avoiding flipping
    pose_flip_ref = pose_ref_index

    # compute background
    if subtract_background:
        try:
            logger.info("Loading background image ...")
            bg = np.load(os.path.join(path_to_file,"videos",filename+'-background.npy'))
        except Exception:
            logger.info("Can't find background image... Calculate background image...")
            bg = background(path_to_file,filename, file_format, save_background=True)

    images = []
    points = []

    for i in pose_list:
        for j in i:
            if j[2] <= 0.8:
                j[0],j[1] = np.nan, np.nan


    for i in pose_list:
        i = interpol_first_rows_nans(i)

    capture = cv.VideoCapture(os.path.join(path_to_file,"videos",filename+file_format))
    if not capture.isOpened():
        raise Exception("Unable to open video file: {0}".format(os.path.join(path_to_file,"videos",filename++file_format)))

    for idx in tqdm.tqdm(range(length), disable=not True, desc='Align frames'):
        try:
            capture.set(1,idx+start+lag)
            ret, frame = capture.read()
            frame = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
            if subtract_background:
                frame = frame - bg
                frame[frame <= 0] = 0
        except Exception:
            logger.info("Couldn't find a frame in capture.read(). #Frame: %d" %idx+start+lag)
            continue

       #Read coordinates and add border
        pose_list_bordered = []

        for i in pose_list:
            pose_list_bordered.append((int(i[idx+start+lag][0]+crop_size[0]),int(i[idx+start+lag][1]+crop_size[1])))

        img = cv.copyMakeBorder(frame, crop_size[1], crop_size[1], crop_size[0], crop_size[0], cv.BORDER_CONSTANT, 0)

        punkte = []
        for i in pose_ref_index:
            coord = []
            coord.append(pose_list_bordered[i][0])
            coord.append(pose_list_bordered[i][1])
            punkte.append(coord)
        punkte = [punkte]
        punkte = np.asarray(punkte)

        #calculate minimal rectangle around snout and tail
        rect = cv.minAreaRect(punkte)

        #change size in rect tuple structure to be equal to crop_size
        lst = list(rect)
        lst[1] = crop_size
        rect = tuple(lst)

        center, size, theta = rect

        #crop image
        out, shifted_points = crop_and_flip(rect, img,pose_list_bordered,pose_flip_ref)

        images.append(out)
        points.append(shifted_points)

    capture.release()
    return images



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 0.1 Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""

import sys
sys.dont_write_bytecode = True

from vame.model.create_training import create_trainset
from vame.model.dataloader import SEQUENCE_DATASET
from vame.model.rnn_vae import train_model
from vame.model.evaluate import evaluate_model




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 0.1 Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""
import sys
sys.dont_write_bytecode = True

from vame.initialize_project import init_new_project
from vame.model import create_trainset
from vame.model import train_model
from vame.model import evaluate_model
from vame.analysis import pose_segmentation
from vame.analysis import motif_videos
from vame.analysis import community
from vame.analysis import community_videos
from vame.analysis import visualization
from vame.analysis import generative_model
from vame.analysis import gif
from vame.util.csv_to_npy import csv_to_numpy
from vame.util.align_egocentrical import egocentric_alignment
from vame.util import model_util
from vame.util import auxiliary




"""
Variational Animal Motion Embedding 1.0-alpha Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""

import os

import torch
import numpy as np
from pathlib import Path
import matplotlib.pyplot as plt
from sklearn.mixture import GaussianMixture
from vame.schemas.states import GenerativeModelFunctionSchema, save_state
from vame.util.auxiliary import read_config
from vame.logging.logger import VameLogger
from typing import Dict
from vame.util.model_util import load_model
from vame.schemas.project import Parametrizations


logger_config = VameLogger(__name__)
logger = logger_config.logger


def random_generative_samples_motif(
    cfg: dict,
    model: torch.nn.Module,
    latent_vector: np.ndarray,
    labels: np.ndarray,
    n_cluster: int
) -> None:
    """Generate random samples for motifs.

    Args:
        cfg (dict): Configuration dictionary.
        model (torch.nn.Module): PyTorch model.
        latent_vector (np.ndarray): Latent vectors.
        labels (np.ndarray): Labels.
        n_cluster (int): Number of clusters.

    Returns:
        None: Plot of generated samples.
    """
    logger.info('Generate random generative samples for motifs...')
    time_window = cfg['time_window']
    for j in range(n_cluster):

        inds=np.where(labels==j)
        motif_latents=latent_vector[inds[0],:]
        gm = GaussianMixture(n_components=10).fit(motif_latents)

        # draw sample from GMM
        density_sample = gm.sample(10)

        # generate image via model decoder
        tensor_sample = torch.from_numpy(density_sample[0]).type('torch.FloatTensor')
        if torch.cuda.is_available():
            tensor_sample = tensor_sample.cuda()
        else:
            tensor_sample = tensor_sample.cpu()

        decoder_inputs = tensor_sample.unsqueeze(2).repeat(1, 1, time_window)
        decoder_inputs = decoder_inputs.permute(0,2,1)

        image_sample = model.decoder(decoder_inputs, tensor_sample)
        recon_sample = image_sample.cpu().detach().numpy()

        fig, axs = plt.subplots(2,5)
        for i in range(5):
            axs[0,i].plot(recon_sample[i,...])
            axs[1,i].plot(recon_sample[i+5,...])
        plt.suptitle('Generated samples for motif '+str(j))
        return fig

def random_generative_samples(cfg: dict, model: torch.nn.Module, latent_vector: np.ndarray) -> None:
    """Generate random generative samples.

    Args:
        cfg (dict): Configuration dictionary.
        model (torch.nn.Module): PyTorch model.
        latent_vector (np.ndarray): Latent vectors.

    Returns:
        None
    """
    logger.info('Generate random generative samples...')
    # Latent sampling and generative model
    time_window = cfg['time_window']
    gm = GaussianMixture(n_components=10).fit(latent_vector)

    # draw sample from GMM
    density_sample = gm.sample(10)

    # generate image via model decoder
    tensor_sample = torch.from_numpy(density_sample[0]).type('torch.FloatTensor')
    if torch.cuda.is_available():
        tensor_sample = tensor_sample.cuda()
    else:
        tensor_sample = tensor_sample.cpu()

    decoder_inputs = tensor_sample.unsqueeze(2).repeat(1, 1, time_window)
    decoder_inputs = decoder_inputs.permute(0,2,1)

    image_sample = model.decoder(decoder_inputs, tensor_sample)
    recon_sample = image_sample.cpu().detach().numpy()

    fig, axs = plt.subplots(2,5)
    for i in range(5):
        axs[0,i].plot(recon_sample[i,...])
        axs[1,i].plot(recon_sample[i+5,...])
    plt.suptitle('Generated samples')
    return fig


def random_reconstruction_samples(cfg: dict, model: torch.nn.Module, latent_vector: np.ndarray) -> None:
    """Generate random reconstruction samples.

    Args:
        cfg (dict): Configuration dictionary.
        model (torch.nn.Module): PyTorch model to use.
        latent_vector (np.ndarray): Latent vectors.

    Returns:
        None
    """
    logger.info('Generate random reconstruction samples...')
    # random samples for reconstruction
    time_window = cfg['time_window']

    rnd = np.random.choice(latent_vector.shape[0], 10)
    tensor_sample = torch.from_numpy(latent_vector[rnd]).type('torch.FloatTensor')
    if torch.cuda.is_available():
        tensor_sample = tensor_sample.cuda()
    else:
        tensor_sample = tensor_sample.cpu()

    decoder_inputs = tensor_sample.unsqueeze(2).repeat(1, 1, time_window)
    decoder_inputs = decoder_inputs.permute(0,2,1)

    image_sample = model.decoder(decoder_inputs, tensor_sample)
    recon_sample = image_sample.cpu().detach().numpy()

    fig, axs = plt.subplots(2,5)
    for i in range(5):
        axs[0,i].plot(recon_sample[i,...])
        axs[1,i].plot(recon_sample[i+5,...])
    plt.suptitle('Reconstructed samples')
    return fig


def visualize_cluster_center(cfg: dict, model: torch.nn.Module, cluster_center: np.ndarray) -> None:
    """Visualize cluster centers.

    Args:
        cfg (dict): Configuration dictionary.
        model (torch.nn.Module): PyTorch model.
        cluster_center (np.ndarray): Cluster centers.

    Returns:
        None
    """
    #Cluster Center
    logger.info('Visualize cluster center...')
    time_window = cfg['time_window']
    animal_centers = cluster_center

    tensor_sample = torch.from_numpy(animal_centers).type('torch.FloatTensor')
    if torch.cuda.is_available():
        tensor_sample = tensor_sample.cuda()
    else:
        tensor_sample = tensor_sample.cpu()
    decoder_inputs = tensor_sample.unsqueeze(2).repeat(1, 1, time_window)
    decoder_inputs = decoder_inputs.permute(0,2,1)

    image_sample = model.decoder(decoder_inputs, tensor_sample)
    recon_sample = image_sample.cpu().detach().numpy()

    num = animal_centers.shape[0]
    b = int(np.ceil(num / 5))

    fig, axs = plt.subplots(5,b)
    idx = 0
    for k in range(5):
        for i in range(b):
            axs[k,i].plot(recon_sample[idx,...])
            axs[k,i].set_title("Cluster %d" %idx)
            idx +=1
    return fig


@save_state(model=GenerativeModelFunctionSchema)
def generative_model(config: str,  parametrization: Parametrizations, mode: str = "sampling", save_logs: bool = False) -> Dict[str, plt.Figure]:
    """Generative model.

    Args:
        config (str): Path to the configuration file.
        mode (str, optional): Mode for generating samples. Defaults to "sampling".

    Returns:
        Dict[str, plt.Figure]: Plots of generated samples for each parametrization.
    """
    try:
        config_file = Path(config).resolve()
        cfg = read_config(config_file)
        if save_logs:
            logs_path = Path(cfg['project_path']) / "logs" / 'generative_model.log'
            logger_config.add_file_handler(logs_path)
        logger.info(f'Running generative model with mode {mode}...')
        model_name = cfg['model_name']
        n_cluster = cfg['n_cluster']
        parametrizations = cfg['parametrizations']

        files = []
        if cfg['all_data'] == 'No':
            all_flag = input("Do you want to write motif videos for your entire dataset? \n"
                        "If you only want to use a specific dataset type filename: \n"
                        "yes/no/filename ")
        else:
            all_flag = 'yes'

        if all_flag == 'yes' or all_flag == 'Yes':
            for file in cfg['video_sets']:
                files.append(file)

        elif all_flag == 'no' or all_flag == 'No':
            for file in cfg['video_sets']:
                use_file = input("Do you want to quantify " + file + "? yes/no: ")
                if use_file == 'yes':
                    files.append(file)
                if use_file == 'no':
                    continue
        else:
            files.append(all_flag)


        model = load_model(cfg, model_name, fixed=False)

        for file in files:
            path_to_file=os.path.join(cfg['project_path'],"results",file,model_name, parametrization + '-' +str(n_cluster),"")

            if mode == "sampling":
                latent_vector = np.load(os.path.join(path_to_file,'latent_vector_'+file+'.npy'))
                return random_generative_samples(cfg, model, latent_vector)

            if mode == "reconstruction":
                latent_vector = np.load(os.path.join(path_to_file,'latent_vector_'+file+'.npy'))
                return random_reconstruction_samples(cfg, model, latent_vector)

            if mode == "centers":
                if parametrization != 'kmeans':
                    raise ValueError(f"Parametrization {parametrization} not supported for cluster center visualization.")
                cluster_center = np.load(os.path.join(path_to_file,'cluster_center_'+file+'.npy'))
                return visualize_cluster_center(cfg, model, cluster_center)

            if mode == "motifs":
                latent_vector = np.load(os.path.join(path_to_file,'latent_vector_'+file+'.npy'))
                labels = np.load(os.path.join(path_to_file,"",str(n_cluster)+'_' + parametrization + '_label_'+file+'.npy'))
                return random_generative_samples_motif(cfg, model, latent_vector,labels,n_cluster)
    except Exception as e:
        logger.exception(str(e))
        raise
    finally:
        logger_config.remove_file_handler()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 1.0-alpha Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""

import os
import numpy as np
import pandas as pd

from pathlib import Path
from vame.util.auxiliary import read_config
from typing import Tuple
from vame.schemas.states import CsvToNumpyFunctionSchema, save_state
from vame.logging.logger import VameLogger
from vame.util.data_manipulation import interpol_first_rows_nans


logger_config = VameLogger(__name__)
logger = logger_config.logger



@save_state(model=CsvToNumpyFunctionSchema)
def csv_to_numpy(config: str, save_logs=False) -> None:
    """Converts a pose-estimation.csv file to a numpy array. Note that this code is only useful for data which is a priori egocentric, i.e. head-fixed
    or otherwise restrained animals.

    Raises:
        ValueError: If the config.yaml file indicates that the data is not egocentric.
    """
    try:
        config_file = Path(config).resolve()
        cfg = read_config(config_file)

        if save_logs:
            log_path = Path(cfg['project_path']) / 'logs' / 'csv_to_numpy.log'
            logger_config.add_file_handler(log_path)


        path_to_file = cfg['project_path']
        filename = cfg['video_sets']
        confidence = cfg['pose_confidence']
        if not cfg['egocentric_data']:
            raise ValueError("The config.yaml indicates that the data is not egocentric. Please check the parameter egocentric_data")

        for file in filename:
            # Read in your .csv file, skip the first two rows and create a numpy array
            data = pd.read_csv(os.path.join(path_to_file,"videos","pose_estimation",file+'.csv'), skiprows = 3, header=None)
            data_mat = pd.DataFrame.to_numpy(data)
            data_mat = data_mat[:,1:]

            pose_list = []

            # get the number of bodyparts, their x,y-position and the confidence from DeepLabCut
            for i in range(int(data_mat.shape[1]/3)):
                pose_list.append(data_mat[:,i*3:(i+1)*3])

            # find low confidence and set them to NaN
            for i in pose_list:
                for j in i:
                    if j[2] <= confidence:
                        j[0],j[1] = np.nan, np.nan

            # interpolate NaNs
            for i in pose_list:
                i = interpol_first_rows_nans(i)

            positions = np.concatenate(pose_list, axis=1)
            final_positions = np.zeros((data_mat.shape[0], int(data_mat.shape[1]/3)*2))

            jdx = 0
            idx = 0
            for i in range(int(data_mat.shape[1]/3)):
                final_positions[:,idx:idx+2] = positions[:,jdx:jdx+2]
                jdx += 3
                idx += 2

            # save the final_positions array with np.save()
            np.save(os.path.join(path_to_file,'data',file,file+"-PE-seq.npy"), final_positions.T)
            logger.info("conversion from DeepLabCut csv to numpy complete...")

        logger.info("Your data is now in right format and you can call vame.create_trainset()")
    except Exception as e:
        logger.exception(f"{e}")
        raise e
    finally:
        logger_config.remove_file_handler()



from functools import wraps
from pydantic import BaseModel, Field
from typing import Optional, Dict
from pathlib import Path
import json
from enum import Enum
from vame.schemas.project import Parametrizations

class StatesEnum(str, Enum):
    success = 'success'
    failed = 'failed'
    running = 'running'
    aborted = 'aborted'

class GenerativeModelModeEnum(str, Enum):
    sampling = 'sampling'
    reconstruction = 'reconstruction'
    centers = 'centers'
    motifs = 'motifs'

class BaseStateSchema(BaseModel):
    config: str = Field(title='Configuration file path')
    execution_state: StatesEnum | None = Field(title='Method execution state', default=None)



class EgocentricAlignmentFunctionSchema(BaseStateSchema):
    pose_ref_index: list = Field(title='Pose reference index', default=[0, 5])
    crop_size: tuple = Field(title='Crop size', default=(300, 300))
    use_video: bool = Field(title='Use video', default=False)
    video_format: str = Field(title='Video format', default='.mp4')
    check_video: bool = Field(title='Check video', default=False)


class CsvToNumpyFunctionSchema(BaseStateSchema):
    ...


class CreateTrainsetFunctionSchema(BaseStateSchema):
    pose_ref_index: Optional[list] = Field(title='Pose reference index', default=None)
    check_parameter: bool = Field(title='Check parameter', default=False)


class TrainModelFunctionSchema(BaseStateSchema):
    ...


class EvaluateModelFunctionSchema(BaseStateSchema):
    use_snapshots: bool = Field(title='Use snapshots', default=False)


class PoseSegmentationFunctionSchema(BaseStateSchema):
    ...

class MotifVideosFunctionSchema(BaseStateSchema):
    videoType: str = Field(title='Type of video', default='.mp4')
    parametrization: Parametrizations = Field(title='Parametrization')
    output_video_type: str = Field(title='Type of output video', default='.mp4')


class CommunityFunctionSchema(BaseStateSchema):
    cohort: bool = Field(title='Cohort', default=True)
    parametrization: Parametrizations = Field(title='Parametrization')
    cut_tree: int | None = Field(title='Cut tree', default=None)


class CommunityVideosFunctionSchema(BaseStateSchema):
    parametrization: Parametrizations = Field(title='Parametrization')
    videoType: str = Field(title='Type of video', default='.mp4')


class VisualizationFunctionSchema(BaseStateSchema):
    parametrization: Parametrizations = Field(title='Parametrization')
    label: Optional[str] = Field(title='Type of labels to visualize', default=None)

class GenerativeModelFunctionSchema(BaseStateSchema):
    parametrization: Parametrizations = Field(title='Parametrization')
    mode: GenerativeModelModeEnum = Field(title='Mode for generating samples', default=GenerativeModelModeEnum.sampling)

class VAMEPipelineStatesSchema(BaseModel):
    egocentric_alignment: Optional[EgocentricAlignmentFunctionSchema | Dict] = Field(title='Egocentric alignment', default={})
    csv_to_numpy: Optional[CsvToNumpyFunctionSchema | Dict] = Field(title='CSV to numpy', default={})
    create_trainset: Optional[CreateTrainsetFunctionSchema | Dict] = Field(title='Create trainset', default={})
    train_model: Optional[TrainModelFunctionSchema | Dict] = Field(title='Train model', default={})
    evaluate_model: Optional[EvaluateModelFunctionSchema | Dict] = Field(title='Evaluate model', default={})
    pose_segmentation: Optional[PoseSegmentationFunctionSchema | Dict] = Field(title='Pose segmentation', default={})
    motif_videos: Optional[MotifVideosFunctionSchema | Dict] = Field(title='Motif videos', default={})
    community: Optional[CommunityFunctionSchema | Dict] = Field(title='Community', default={})
    community_videos: Optional[CommunityVideosFunctionSchema | Dict] = Field(title='Community videos', default={})
    visualization: Optional[VisualizationFunctionSchema | Dict] = Field(title='Visualization', default={})
    generative_model: Optional[GenerativeModelFunctionSchema | Dict] = Field(title='Generative model', default={})


def _save_state(model: BaseModel, function_name: str, state: StatesEnum) -> None:
    """
    Save the state of the function to the project states json file.
    """
    config_file_path = Path(model.config)
    project_path = config_file_path.parent
    states_file_path = project_path / 'states/states.json'

    with open(states_file_path, 'r') as f:
        states = json.load(f)

    pipeline_states = VAMEPipelineStatesSchema(**states)
    model.execution_state = state
    setattr(pipeline_states, function_name, model.model_dump())

    with open(states_file_path, 'w') as f:
        json.dump(pipeline_states.model_dump(), f, indent=4)


def save_state(model: BaseModel):
    """
    Decorator responsible for validating function arguments using pydantic and
    saving the state of the called function to the project states json file.
    """
    def decorator(func: callable):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Create an instance of the Pydantic model using provided args and kwargs
            function_name =  func.__name__
            attribute_names = list(model.model_fields.keys())

            kwargs_dict = {}
            for attr in attribute_names:
                if attr == 'execution_state':
                    kwargs_dict[attr] = 'running'
                    continue
                kwargs_dict[attr] = kwargs.get(attr, model.model_fields[attr].default)

            # Override with positional arguments
            for i, arg in enumerate(args):
                kwargs_dict[attribute_names[i]] = arg
            # Validate function args and kwargs using the Pydantic model.
            kwargs_model = model(**kwargs_dict)
            _save_state(kwargs_model, function_name, state=StatesEnum.running)
            try:
                func_output = func(*args, **kwargs)
                _save_state(kwargs_model, function_name, state=StatesEnum.success)
                return func_output
            except Exception as e:
                _save_state(kwargs_model, function_name, state=StatesEnum.failed)
                raise e
            except KeyboardInterrupt as e:
                _save_state(kwargs_model, function_name, state=StatesEnum.aborted)
                raise e
        return wrapper
    return decorator


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 1.0-alpha Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""

import os
import tqdm
import umap
import numpy as np
from pathlib import Path
import matplotlib
from matplotlib import pyplot as plt
from matplotlib.gridspec import GridSpec
from vame.util.auxiliary import read_config
from vame.util.gif_pose_helper import get_animal_frames
from typing import List, Tuple
from vame.logging.logger import VameLogger
from vame.schemas.project import Parametrizations



logger_config = VameLogger(__name__)
logger = logger_config.logger


def create_video(
    path_to_file: str,
    file: str,
    embed: np.ndarray,
    clabel: np.ndarray,
    frames: List[np.ndarray],
    start: int,
    length: int,
    max_lag: int,
    num_points: int
) -> None:
    """Create video frames for the given embedding.

    Args:
        path_to_file (str): Path to the file.
        file (str): File name.
        embed (np.ndarray): Embedding array.
        clabel (np.ndarray): Cluster labels.
        frames (List[np.ndarray]): List of frames.
        start (int): Starting index.
        length (int): Length of the video.
        max_lag (int): Maximum lag.
        num_points (int): Number of points.

    Returns:
        None
    """
    # set matplotlib colormap
    cmap = matplotlib.cm.gray
    cmap_reversed = plt.get_cmap('gray_r')

    # this here generates every frame for your gif. The gif is lastly created by using ImageJ
    # the embed variable is my umap embedding, which is for the 2D case a 2xn dimensional vector
    fig = plt.figure()
    spec = GridSpec(ncols=2, nrows=1, width_ratios=[6, 3])
    ax1 = fig.add_subplot(spec[0])
    ax2 = fig.add_subplot(spec[1])
    ax2.axis('off')
    ax2.grid(False)
    lag = 0
    for i in tqdm.tqdm(range(length)):
        if i > max_lag:
            lag = i - max_lag
        ax1.cla()
        ax1.axis('off')
        ax1.grid(False)
        if clabel is not None:
            ax1.scatter(embed[:num_points,0], embed[:num_points,1], c=clabel[:num_points], cmap='Spectral', s=1, alpha=0.4)
        else:
            ax1.scatter(embed[:num_points,0], embed[:num_points,1], s=1, alpha=0.4)

        ax1.set_aspect('equal', 'datalim')
        ax1.plot(embed[start+lag:start+i,0], embed[start+lag:start+i,1],'.b-',alpha=.6, linewidth=2, markersize=4)
        ax1.plot(embed[start+i,0], embed[start+i,1], 'gx', markersize=4)
        frame = frames[i]
        ax2.imshow(frame, cmap=cmap_reversed)
        # ax2.set_title("Motif %d,\n Community: %s" % (lbl, motifs[lbl]), fontsize=10)
        fig.savefig(os.path.join(path_to_file,"gif_frames",file+'gif_%d.png') %i)


def gif(
    config: str,
    pose_ref_index: int,
    parametrization: Parametrizations,
    subtract_background: bool = True,
    start: int | None = None,
    length: int = 500,
    max_lag: int = 30,
    label: str = 'community',
    file_format: str = '.mp4',
    crop_size: Tuple[int, int] = (300,300)
) -> None:
    """Create a GIF from the given configuration.

    Args:
        config (str): Path to the configuration file.
        pose_ref_index (int): Pose reference index.
        subtract_background (bool, optional): Whether to subtract background. Defaults to True.
        start (int, optional): Starting index. Defaults to None.
        length (int, optional): Length of the video. Defaults to 500.
        max_lag (int, optional): Maximum lag. Defaults to 30.
        label (str, optional): Label type [None, community, motif]. Defaults to 'community'.
        file_format (str, optional): File format. Defaults to '.mp4'.
        crop_size (Tuple[int, int], optional): Crop size. Defaults to (300,300).

    Returns:
        None
    """

    config_file = Path(config).resolve()
    cfg = read_config(config_file)
    model_name = cfg['model_name']
    n_cluster = cfg['n_cluster']

    if parametrization not in cfg['parametrizations']:
        raise ValueError("Parametrization not found in config")

    files = []
    if cfg['all_data'] == 'No':
        all_flag = input("Do you want to write motif videos for your entire dataset? \n"
                     "If you only want to use a specific dataset type filename: \n"
                     "yes/no/filename ")
    else:
        all_flag = 'yes'

    if all_flag == 'yes' or all_flag == 'Yes':
        for file in cfg['video_sets']:
            files.append(file)

    elif all_flag == 'no' or all_flag == 'No':
        for file in cfg['video_sets']:
            use_file = input("Do you want to quantify " + file + "? yes/no: ")
            if use_file == 'yes':
                files.append(file)
            if use_file == 'no':
                continue
    else:
        files.append(all_flag)


    for file in files:
        path_to_file=os.path.join(cfg['project_path'],"results",file,model_name, parametrization+'-'+str(n_cluster),"")
        if not os.path.exists(os.path.join(path_to_file,"gif_frames")):
            os.mkdir(os.path.join(path_to_file,"gif_frames"))

        embed = np.load(os.path.join(path_to_file,"community","umap_embedding_"+file+'.npy'))

        try:
            embed = np.load(os.path.join(path_to_file,"","community","","umap_embedding_"+file+".npy"))
            num_points = cfg['num_points']
            if num_points > embed.shape[0]:
                num_points = embed.shape[0]
        except Exception:
            logger.info(f"Compute embedding for file {file}")
            reducer = umap.UMAP(n_components=2, min_dist=cfg['min_dist'], n_neighbors=cfg['n_neighbors'],
                    random_state=cfg['random_state'])

            latent_vector = np.load(os.path.join(path_to_file,"",'latent_vector_'+file+'.npy'))

            num_points = cfg['num_points']
            if num_points > latent_vector.shape[0]:
                num_points = latent_vector.shape[0]
            logger.info("Embedding %d data points.." %num_points)

            embed = reducer.fit_transform(latent_vector[:num_points,:])
            np.save(os.path.join(path_to_file,"community","umap_embedding_"+file+'.npy'), embed)

        if label == "motif":
            umap_label = np.load(os.path.join(path_to_file,str(n_cluster)+"_" + parametrization + "_label_"+file+'.npy'))
        elif label == "community":
            umap_label = np.load(os.path.join(path_to_file,"community","community_label_"+file+'.npy'))
        elif label is None:
            umap_label = None

        if start is None:
            start = np.random.choice(embed[:num_points].shape[0]-length)
        else:
            start = start

        frames = get_animal_frames(cfg, file, pose_ref_index, start, length, subtract_background, file_format, crop_size)
        create_video(path_to_file, file, embed, umap_label, frames, start, length, max_lag, num_points)






























#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 1.0-alpha Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""


import numpy as np
import networkx as nx
import random
from matplotlib import pyplot as plt
from typing import Dict, List, Tuple


def hierarchy_pos(
    G: nx.Graph,
    root: str | None = None,
    width: float = 0.5,
    vert_gap: float = 0.2,
    vert_loc: float = 0,
    xcenter: float = 0.5
) -> Dict[str, Tuple[float, float]]:
    """
    Positions nodes in a tree-like layout.
    Ref: From Joel's answer at https://stackoverflow.com/a/29597209/2966723.

    Args:
        G (nx.Graph): The input graph. Must be a tree.
        root (str, optional): The root node of the tree. If None, the function selects a root node based on graph type.
        width (float, optional): The horizontal space assigned to each level.
        vert_gap (float, optional): The vertical gap between levels.
        vert_loc (float, optional): The vertical location of the root node.
        xcenter (float, optional): The horizontal location of the root node.

    Returns:
        Dict[str, Tuple[float, float]]: A dictionary mapping node names to their positions (x, y).
    """

    if not nx.is_tree(G):
        raise TypeError('cannot use hierarchy_pos on a graph that is not a tree')

    if root is None:
        if isinstance(G, nx.DiGraph):
            root = next(iter(nx.topological_sort(G)))  #allows back compatibility with nx version 1.11
        else:
            root = random.choice(list(G.nodes))

    def _hierarchy_pos(G, root, width=1., vert_gap = 0.2, vert_loc = 0, xcenter = 0.5, pos = None, parent = None):
        if pos is None:
            pos = {root:(xcenter,vert_loc)}
        else:
            pos[root] = (xcenter, vert_loc)
        children = list(G.neighbors(root))
        if not isinstance(G, nx.DiGraph) and parent is not None:
            children.remove(parent)
        if len(children)!=0:
            dx = width/len(children)
            nextx = xcenter - width/2 - dx/2
            for child in children:
                nextx += dx
                pos = _hierarchy_pos(G,child, width = dx, vert_gap = vert_gap,
                                    vert_loc = vert_loc-vert_gap, xcenter=nextx,
                                    pos=pos, parent = root)
        return pos


    return _hierarchy_pos(G, root, width, vert_gap, vert_loc, xcenter)


def merge_func(
    transition_matrix: np.ndarray,
    n_cluster: int,
    motif_norm: np.ndarray,
    merge_sel: int
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Merge nodes in a graph based on a selection criterion.

    Args:
        transition_matrix (np.ndarray): The transition matrix of the graph.
        n_cluster (int): The number of clusters.
        motif_norm (np.ndarray): The normalized motif matrix.
        merge_sel (int): The merge selection criterion.
            - 0: Merge nodes with highest transition probability.
            - 1: Merge nodes with lowest cost.

    Raises:
        ValueError: If an invalid merge selection criterion is provided.

    Returns:
        Tuple[np.ndarray, np.ndarray]: A tuple containing the merged nodes.
    """

    if merge_sel == 0:
        # merge nodes with highest transition probability
        cost = np.max(transition_matrix)
        merge_nodes = np.where(cost == transition_matrix)

    if merge_sel == 1:

        cost_temp = 100
        for i in range(n_cluster):
            for j in range(n_cluster):
                try:
                    cost = motif_norm[i] + motif_norm[j] / np.abs(transition_matrix[i,j] + transition_matrix[j,i] )
                except ZeroDivisionError:
                    print("Error: Transition probabilities between motif "+str(i)+" and motif "+str(j)+ " are zero.")
                if cost <= cost_temp:
                    cost_temp = cost
                    merge_nodes = (np.array([i]), np.array([j]))
    else:
        raise ValueError("Invalid merge selection criterion. Please select 0 or 1.")

    return merge_nodes


def graph_to_tree(
    motif_usage: np.ndarray,
    transition_matrix: np.ndarray,
    n_cluster: int,
    merge_sel: int = 1
) -> nx.Graph:
    """
    Convert a graph to a tree.

    Args:
        motif_usage (np.ndarray): The motif usage matrix.
        transition_matrix (np.ndarray): The transition matrix of the graph.
        n_cluster (int): The number of clusters.
        merge_sel (int, optional): The merge selection criterion. Defaults to 1.
            - 0: Merge nodes with highest transition probability.
            - 1: Merge nodes with lowest cost.

    Returns:
        nx.Graph: The tree.
    """

    if merge_sel == 1:
        # motif_usage_temp = np.load(path_to_file+'/behavior_quantification/motif_usage.npy')
        motif_usage_temp = motif_usage
        motif_usage_temp_colsum = motif_usage_temp.sum(axis=0)
        motif_norm = motif_usage_temp/motif_usage_temp_colsum
        motif_norm_temp = motif_norm.copy()
    else:
        motif_norm_temp = None

    merging_nodes = []
    hierarchy_nodes = []
    trans_mat_temp = transition_matrix.copy()
    is_leaf = np.ones((n_cluster), dtype='int')
    node_label = []
    leaf_idx = []

    if np.any(transition_matrix.sum(axis=1) == 0):
        temp = np.where(transition_matrix.sum(axis=1)==0)
        reduction = len(temp) + 1
    else:
        reduction = 1

    for i in range(n_cluster-reduction):
        nodes = merge_func(trans_mat_temp, n_cluster, motif_norm_temp, merge_sel)

        if np.size(nodes) >= 2:
            nodes = np.array([nodes[0][0], nodes[1][0]])

        if is_leaf[nodes[0]] == 1:
            is_leaf[nodes[0]] = 0
            node_label.append('leaf_left_'+str(i))
            leaf_idx.append(1)

        elif is_leaf[nodes[0]] == 0:
            node_label.append('h_'+str(i)+'_'+str(nodes[0]))
            leaf_idx.append(0)

        if is_leaf[nodes[1]] == 1:
            is_leaf[nodes[1]] = 0
            node_label.append('leaf_right_'+str(i))
            hierarchy_nodes.append('h_'+str(i)+'_'+str(nodes[1]))
            leaf_idx.append(1)

        elif is_leaf[nodes[1]] == 0:
            node_label.append('h_'+str(i)+'_'+str(nodes[1]))
            hierarchy_nodes.append('h_'+str(i)+'_'+str(nodes[1]))
            leaf_idx.append(0)

        merging_nodes.append(nodes)

        node1_trans_x = trans_mat_temp[nodes[0],:]
        node2_trans_x = trans_mat_temp[nodes[1],:]

        node1_trans_y = trans_mat_temp[:,nodes[0]]
        node2_trans_y = trans_mat_temp[:,nodes[1]]

        new_node_trans_x = node1_trans_x + node2_trans_x
        new_node_trans_y = node1_trans_y + node2_trans_y

        trans_mat_temp[nodes[1],:] = new_node_trans_x
        trans_mat_temp[:,nodes[1]] = new_node_trans_y

        trans_mat_temp[nodes[0],:] = 0
        trans_mat_temp[:,nodes[0]] = 0

        trans_mat_temp[nodes[1],nodes[1]] = 0

        if merge_sel == 1:
            motif_norm_1 = motif_norm_temp[nodes[0]]
            motif_norm_2 = motif_norm_temp[nodes[1]]

            new_motif = motif_norm_1 + motif_norm_2

            motif_norm_temp[nodes[0]] = 0
            motif_norm_temp[nodes[1]] = 0

            motif_norm_temp[nodes[1]] = new_motif

    merge = np.array(merging_nodes)

    T = nx.Graph()

    T.add_node('Root')
    node_dict = {}

    if leaf_idx[-1] == 0:
        temp_node = 'h_'+str(merge[-1,1])+'_'+str(28)
        T.add_edge(temp_node, 'Root')
        node_dict[merge[-1,1]] = temp_node

    if leaf_idx[-1] == 1:
        T.add_edge(merge[-1,1], 'Root')

    if leaf_idx[-2] == 0:
        temp_node = 'h_'+str(merge[-1,0])+'_'+str(28)
        T.add_edge(temp_node, 'Root')
        node_dict[merge[-1,0]] = temp_node

    if leaf_idx[-2] == 1:
        T.add_edge(merge[-1,0], 'Root')

    idx = len(leaf_idx)-3

    if np.any(transition_matrix.sum(axis=1) == 0):
        temp = np.where(transition_matrix.sum(axis=1)==0)
        reduction = len(temp) + 2
    else:
        reduction = 2

    for i in range(n_cluster-reduction)[::-1]:

        if leaf_idx[idx-1] == 1:
            if merge[i,1] in node_dict:
                T.add_edge(merge[i,0], node_dict[merge[i,1]])
            else:
                T.add_edge(merge[i,0], temp_node)

        if leaf_idx[idx] == 1:
            if merge[i,1] in node_dict:
                T.add_edge(merge[i,1], node_dict[merge[i,1]])
            else:
                T.add_edge(merge[i,1], temp_node)

        if leaf_idx[idx] == 0:
            new_node = 'h_'+str(merge[i,1])+'_'+str(i)
            if merge[i,1] in node_dict:
                T.add_edge(node_dict[merge[i,1]], new_node)
            else:
                T.add_edge(temp_node, new_node)

            if leaf_idx[idx-1] == 1:
                temp_node = new_node
                node_dict[merge[i,1]] = new_node
            else:
                new_node_2 = 'h_'+str(merge[i,0])+'_'+str(i)
                T.add_edge(node_dict[merge[i,1]], new_node_2)
                node_dict[merge[i,1]] = new_node
                node_dict[merge[i,0]] = new_node_2

        elif leaf_idx[idx-1] == 0:
            new_node = 'h_'+str(merge[i,0])+'_'+str(i)
            if merge[i,1] in node_dict:
                T.add_edge(node_dict[merge[i,1]], new_node)
            else:
                T.add_edge(temp_node, new_node)
            node_dict[merge[i,0]] = new_node

            if leaf_idx[idx] == 1:
                temp_node = new_node
            else:
                new_node = 'h_'+str(merge[i,1])+'_'+str(i)
                T.add_edge(temp_node, new_node)
                node_dict[merge[i,1]] = new_node
                temp_node = new_node

        idx -= 2

    return T

def draw_tree(T: nx.Graph) -> None:
    """
    Draw a tree.

    Args:
        T (nx.Graph): The tree to be drawn.

    Returns:
        None
    """
    # pos = nx.drawing.layout.fruchterman_reingold_layout(T)
    pos = hierarchy_pos(T,'Root',width=.5, vert_gap = 0.1, vert_loc = 0, xcenter = 50)
    fig = plt.figure(2)
    nx.draw_networkx(T, pos)
    figManager = plt.get_current_fig_manager()
    #figManager.window.showMaximized()

def _traverse_tree_cutline(
    T: nx.Graph,
    node: List[str],
    traverse_list: List[str],
    cutline: int,
    level: int,
    community_bag: List[List[str]],
    community_list: List[str] = None
) -> List[List[str]]:
    """
    Helper function for tree traversal with a cutline.

    Args:
        T (nx.Graph): The tree to be traversed.
        node (List[str]): Current node being traversed.
        traverse_list (List[str]): List of traversed nodes.
        cutline (int): The cutline level.
        level (int): The current level in the tree.
        community_bag (List[List[str]]): List of community bags.
        community_list (List[str], optional): List of nodes in the current community bag.

    Returns:
        List[List[str]]: List of lists community bags.
    """
    cmap = plt.get_cmap("tab10")
    traverse_list.append(node[0])
    if community_list is not None and type(node[0]) is not str:
        community_list.append(node[0])
    children = list(T.neighbors(node[0]))

    if len(children) == 3:
#        print(children)
        for child in children:
            if child in traverse_list:
#                    print(child)
                children.remove(child)

    if len(children) > 1:
        if nx.shortest_path_length(T,'Root',node[0])==cutline:
            #create new list
            traverse_list1 = []
            traverse_list2 = []
            community_bag = _traverse_tree_cutline(T, [children[0]], traverse_list, cutline, level+1, community_bag, traverse_list1)
            community_bag = _traverse_tree_cutline(T, [children[1]], traverse_list, cutline, level+1, community_bag, traverse_list2)
            joined_list=traverse_list1+traverse_list2
            community_bag.append(joined_list)
            if type(node[0]) is not str: #append itself
                community_bag.append([node[0]])
        else:
             community_bag = _traverse_tree_cutline(T, [children[0]], traverse_list, cutline, level+1, community_bag, community_list)
             community_bag = _traverse_tree_cutline(T, [children[1]], traverse_list, cutline, level+1, community_bag, community_list)

    return  community_bag


def traverse_tree_cutline(T: nx.Graph, root_node: str | None = None, cutline: int = 2) -> List[List[str]]:
    """
    Traverse a tree with a cutline and return the community bags.

    Args:
        T (nx.Graph): The tree to be traversed.
        root_node (str, optional): The root node of the tree. If None, traversal starts from the root.
        cutline (int, optional): The cutline level.

    Returns:
        List[List[str]]: List of community bags.
    """
    if root_node == None:
        node=['Root']
    else:
        node=[root_node]
    traverse_list = []
    color_map = []
    community_bag=[]
    level = 0
    community_bag = _traverse_tree_cutline(T, node, traverse_list,cutline, level, color_map,community_bag)

    return community_bag




import os
import yaml
import ruamel.yaml
from pathlib import Path
from typing import Tuple
import torch
from vame.logging.logger import VameLogger
from vame.model.rnn_model import RNN_VAE


logger_config = VameLogger(__name__)
logger = logger_config.logger

def load_model(cfg: dict, model_name: str, fixed: bool = True) -> RNN_VAE:
    """Load the VAME model.

    Args:
        cfg (dict): Configuration dictionary.
        model_name (str): Name of the model.
        fixed (bool): Fixed or variable length sequences.

    Returns:
        RNN_VAE: Loaded VAME model.
    """
    # load Model
    ZDIMS = cfg['zdims']
    FUTURE_DECODER = cfg['prediction_decoder']
    TEMPORAL_WINDOW = cfg['time_window']*2
    FUTURE_STEPS = cfg['prediction_steps']
    NUM_FEATURES = cfg['num_features']

    if not fixed:
        NUM_FEATURES = NUM_FEATURES - 2
    hidden_size_layer_1 = cfg['hidden_size_layer_1']
    hidden_size_layer_2 = cfg['hidden_size_layer_2']
    hidden_size_rec = cfg['hidden_size_rec']
    hidden_size_pred = cfg['hidden_size_pred']
    dropout_encoder = cfg['dropout_encoder']
    dropout_rec = cfg['dropout_rec']
    dropout_pred = cfg['dropout_pred']
    softplus = cfg['softplus']


    logger.info('Loading model... ')

    model = RNN_VAE(
        TEMPORAL_WINDOW,
        ZDIMS,
        NUM_FEATURES,
        FUTURE_DECODER,
        FUTURE_STEPS,
        hidden_size_layer_1,
        hidden_size_layer_2,
        hidden_size_rec,
        hidden_size_pred,
        dropout_encoder,
        dropout_rec,
        dropout_pred,
        softplus
    )
    if torch.cuda.is_available():
        model = model.cuda()
    else:
        model = model.cpu()

    model.load_state_dict(torch.load(os.path.join(cfg['project_path'],'model','best_model',model_name+'_'+cfg['Project']+'.pkl')))
    model.eval()

    return model



import logging
from pathlib import Path
import io
from datetime import datetime


class VameLogger:
    LOG_FORMAT = (
        "%(asctime)-15s.%(msecs)d %(levelname)-5s --- [%(threadName)s]"
        " %(name)-15s : %(lineno)d : %(message)s"
    )
    LOG_DATE_FORMAT = '%Y-%m-%d %H:%M:%S'

    def __init__(self, base_name: str, file_path: str | None = None, log_level: int = logging.INFO):
        self.log_level = log_level
        self.file_handler = None
        logging.basicConfig(
            level=log_level,
            format=self.LOG_FORMAT,
            datefmt=self.LOG_DATE_FORMAT
        )
        self.logger = logging.getLogger(f'{base_name}')
        if self.logger.hasHandlers():
            self.logger.handlers.clear()

        self.logger.setLevel(self.log_level)
        # Stream handler for logging to stdout
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(
            logging.Formatter(self.LOG_FORMAT, self.LOG_DATE_FORMAT)
        )
        self.logger.addHandler(stream_handler)
        self.logger.propagate = False

        if file_path is not None:
            self.add_file_handler(file_path)

    def add_file_handler(self, file_path: str):
        # File handler for logging to a file
        #if not Path(file_path).parent.exists():

        line_break = '\n'
        if not Path(file_path).exists():
            Path(file_path).parent.mkdir(parents=True, exist_ok=True)
            line_break = ''

        if self.file_handler is not None:
            self.file_handler.close()
            self.logger.removeHandler(self.file_handler)

        handler_datetime = datetime.now()
        handler_datetime = handler_datetime.strftime("%Y-%m-%d %H-%M-%S")
        with open(file_path, "a") as f:
            f.write(f"{line_break}[LOGGING STARTED AT: {handler_datetime}]")

        self.file_handler = logging.FileHandler(file_path, mode='a')
        self.file_handler.setFormatter(logging.Formatter(self.LOG_FORMAT, self.LOG_DATE_FORMAT))
        self.logger.addHandler(self.file_handler)

    def remove_file_handler(self):
        if self.file_handler:
            self.file_handler.close()
            self.logger.removeHandler(self.file_handler)
            self.file_handler = None



class TqdmToLogger(io.StringIO):
    """
    Output stream for TQDM which will output to logger module instead of
    the StdOut.
    """

    logger = None
    level = None
    buf = ""

    def __init__(self, logger, level=None):
        super(TqdmToLogger, self).__init__()
        self.logger = logger
        self.level = level or logging.INFO

    def write(self, buf):
        self.buf = buf.strip("\r\n\t ")

    def flush(self):
        self.logger.log(self.level, self.buf)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Variational Animal Motion Embedding 0.1 Toolbox
© K. Luxem & P. Bauer, Department of Cellular Neuroscience
Leibniz Institute for Neurobiology, Magdeburg, Germany

https://github.com/LINCellularNeuroscience/VAME
Licensed under GNU General Public License v3.0
"""

import torch
from torch.utils.data.dataset import Dataset
import numpy as np
import os
from vame.logging.logger import VameLogger


class SEQUENCE_DATASET(Dataset):
    def __init__(self, path_to_file: str, data: str, train: bool, temporal_window: int, **kwargs) -> None:
        """Initialize the Sequence Dataset.

        Args:
            path_to_file (str): Path to the dataset files.
            data (str): Name of the data file.
            train (bool): Flag indicating whether it's training data.
            temporal_window (int): Size of the temporal window.

        Returns:
            None
        """
        self.logger_config = kwargs.get('logger_config', VameLogger(__name__))
        self.logger = self.logger_config.logger

        self.temporal_window = temporal_window
        self.X = np.load(path_to_file+data)
        if self.X.shape[0] > self.X.shape[1]:
            self.X=self.X.T

        self.data_points = len(self.X[0,:])

        if train and not os.path.exists(os.path.join(path_to_file,'seq_mean.npy')):
            self.logger.info("Compute mean and std for temporal dataset.")
            self.mean = np.mean(self.X)
            self.std = np.std(self.X)
            np.save(path_to_file+'seq_mean.npy', self.mean)
            np.save(path_to_file+'seq_std.npy', self.std)
        else:
            self.mean = np.load(path_to_file+'seq_mean.npy')
            self.std = np.load(path_to_file+'seq_std.npy')

        if train:
            self.logger.info('Initialize train data. Datapoints %d' %self.data_points)
        else:
            self.logger.info('Initialize test data. Datapoints %d' %self.data_points)

    def __len__(self) -> int:
        """Return the number of data points.

        Returns:
            int: Number of data points.
        """
        return self.data_points

    def __getitem__(self, index: int) -> torch.Tensor:
        """
        Get a normalized sequence at the specified index.

        Args:
            index (int): Index of the item.

        Returns:
            torch.Tensor: Normalized sequence data at the specified index.
        """
        temp_window = self.temporal_window

        nf = self.data_points
        start = np.random.choice(nf-temp_window)
        end = start+temp_window

        sequence = self.X[:,start:end]

        sequence = (sequence-self.mean)/self.std

        return torch.from_numpy(sequence)





